/* Copyright 2018-2020 Pavel Ponec, https://jbook.ponec.net */
package net.ponec.jbook.s70_gomoku;

import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.servlet.ServletException;
import javax.servlet.annotation.WebServlet;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import net.ponec.jbook.tools.WebTools;
import org.ujorm.tools.web.Element;
import org.ujorm.tools.web.Html;
import org.ujorm.tools.web.HtmlElement;

/** Simple board game */
@WebServlet("/go-moku")
public class GoMokuServlet extends HttpServlet {

    /** Logger */
    private static final Logger LOGGER = Logger.getLogger(GoMokuServlet.class.getName());

    /** The number of winning stones */
    public static final int WINNING_STONE_COUNT = 5;

    /** Cascading style sheets */
    private static final String[] CSS = new String[] {"css/board.css", "css/gomoku.css"};

    /** A board service */
    private final BoardService service = BoardService.of();

    /**
     * Handles the HTTP <code>GET</code> or <code>POST</code> method.
     * @param request A servlet request encoded by UTF_8
     * @param response A servlet response
     * @throws ServletException if a servlet-specific error occurs
     * @throws IOException if an I/O error occurs
     */
    @Override
    protected void doGet(HttpServletRequest request, HttpServletResponse response)
            throws ServletException, IOException {
        try (HtmlElement html = HtmlElement.niceOf("Go-moku (simple defensive strategy)", response, CSS)) {
            html.getBody().addHeading(html.getTitle());
            try {
                BoardModel board = service.createBoardModel(request);
                Element form = createFormElement(html.getBody(), board);
                board.getErrorMessage().ifPresent(msg -> form.addSpan("msg").addText(msg));
                service.getWinnerMessage(board).ifPresent(msg -> form.addSpan("msg").addText(msg));
                createMainBoard(form, board);
                createResetButton(form, "New game");
            } catch (Exception e) {
                String msg = "Internal server error caused by " + e.getClass().getSimpleName();
                LOGGER.log(Level.SEVERE, msg, e);
                html.getBody().addSpan("msg").addText(msg + ", see a server log for detail");
            }
            WebTools.addFooter(html, this);
        }
    }

    /** Create a form element including hidden fields */
    private Element createFormElement(final Element parent, BoardModel board) {
        Element result = parent.addForm().setMethod(Html.V_GET);
        result.addInput()
                .setType(Html.V_HIDDEN)
                .setName(BoardService.BOARD_PARAM)
                .setValue(board.exportBoard());
        return result;
    }

    /** Create a clear button element */
    private void createResetButton(Element parent, String title) {
        parent.addAnchor(WebTools.getUrlOfServlet(getClass()), "btn", "btn-primary").addText(title);
    }

    /** Create a board element */
    protected void createMainBoard(Element parent, BoardModel board) {
        final BoardPoint boardPoint = new BoardPoint();
        final Element table = parent.addTable("board");
        for (int y = 0; y < board.getHeight(); y++) {
            final Element rowElement = table.addTableRow();
            for (int x = 0; x < board.getWidth(); x++) {
                final Element td = rowElement.addTableDetail();
                final StoneEnum stone = board.getStone(boardPoint.set(x, y));
                switch(stone) {
                    case NO_STONE:
                        td.addSubmitButton()
                            .setName(BoardService.CELL_PREFIX_PARAM + (x + y * board.getWidth()))
                            .addText("");
                        break;
                    default:
                        td.setClass(stone == StoneEnum.WHITE_STONE ? "w" : "b")
                            .addSpan(boardPoint.equals(board.getLastMove()) ? "last" : null)
                            .addText("");
                }
            }
        }
    }

    /**
     * Handles the HTTP <code>POST</code> method.
     * @param request servlet request
     * @param response servlet response
     * @throws ServletException if a servlet-specific error occurs
     * @throws IOException if an I/O error occurs
     */
    @Override
    protected void doPost(HttpServletRequest request, HttpServletResponse response)
            throws ServletException, IOException {
        doGet(request, response);
    }
}
